from pathlib import Path
from typing import Any, List

from fastapi import FastAPI, Depends, HTTPException, Request
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates

from pydantic import BaseModel
from starlette.responses import FileResponse

from buenavista.adapter import AdapterHandle

app = FastAPI(title="DuckDBT")

BASE_PATH = Path(__file__).resolve().parent
TEMPLATES = Jinja2Templates(directory=str(BASE_PATH / "templates"))

app.mount(
    "/static",
    StaticFiles(directory=BASE_PATH / "static"),
    name="static",
)


def bv_handle() -> AdapterHandle:
    handle = app.bv.adapter.create_handle()
    try:
        yield handle
    finally:
        handle.close()


@app.get("/")
def index(request: Request, handle=Depends(bv_handle)):
    qr = handle.execute_sql(
        "SELECT table_schema, table_name, column_name FROM information_schema.columns ORDER BY table_schema, table_name, ordinal_position"
    )
    if qr.has_results():
        res = {}
        for row in qr.rows():
            relation = f"{row[0]}.{row[1]}"
            if relation not in res:
                res[relation] = []
            res[relation].append(row[2])
    else:
        res = {}
    return TEMPLATES.TemplateResponse(
        "index.html", {"request": request, "relations": res}
    )


class QueryRequest(BaseModel):
    sql: str


class QueryResult(BaseModel):
    columns: List[str]
    rows: List[List[Any]]


@app.post("/api/query", response_model=QueryResult)
def query(q: QueryRequest, handle=Depends(bv_handle)):
    try:
        query_result = handle.execute_sql(q.sql)
        if query_result.has_results():
            columns = [
                query_result.column(i)[0] for i in range(query_result.column_count())
            ]
            return QueryResult(columns=columns, rows=list(query_result.rows()))
        return QueryResult(columns=[], rows=[])
    except Exception as e:
        raise HTTPException(status_code=404, detail=str(e))



from typing import Any, Dict

from buenavista.adapter import Adapter
from buenavista.core import BuenaVistaServer
from dbt.adapters.buenavista.python_extension import DbtPythonRunner


def _duckdb_adapter(config: Dict[str, Any]) -> Adapter:
    from buenavista.backends.duckdb import DuckDBAdapter
    import duckdb

    db = duckdb.connect(config["path"])
    for e in config.get("extensions", []):
        db.install_extension(e)
        db.load_extension(e)
    for key, value in config.get("settings", {}).items():
        db.execute(f"SET {key} = '{value}'")
    return DuckDBAdapter(db)


def _postgres_adapter(config: Dict[str, Any]) -> Adapter:
    from buenavista.backends.postgres import PGAdapter

    kwargs = {}
    for k in ("host", "port", "user", "password", "dbname"):
        if k in config:
            kwargs[k] = config[k]
    return PGAdapter(**kwargs)


def create(
    bv_config: Dict[str, Any], backing_config: Dict[str, Any]
) -> BuenaVistaServer:
    backing_type = backing_config["type"]
    if backing_type == "postgres":
        adapter = _postgres_adapter(backing_config)
    elif backing_type == "duckdb":
        adapter = _duckdb_adapter(backing_config)
    else:
        raise Exception(f"Unsupported backing type for Buena Vista: {backing_type}")

    server_address = (bv_config.get("host"), bv_config.get("port"))
    return BuenaVistaServer(server_address, adapter, extensions=[DbtPythonRunner()])


if __name__ == "__main__":
    # Gratefully copied (with slight modifications) from
    # https://github.com/fal-ai/fal/blob/main/adapter/src/dbt/adapters/fal/impl.py#L24
    import os

    from dbt.config.profile import Profile, read_profile
    from dbt.config.project import Project
    from dbt.config.renderer import ProfileRenderer
    from dbt.config.utils import parse_cli_vars
    from dbt.main import _build_base_subparser
    from dbt import flags

    # includes vars, profile, target
    parser = _build_base_subparser()
    args, unknown = parser.parse_known_args()

    # dbt-core does os.chdir(project_dir) before reaching this location
    # from https://github.com/dbt-labs/dbt-core/blob/73116fb816498c4c45a01a2498199465202ec01b/core/dbt/task/base.py#L186
    project_root = os.getcwd()

    # from https://github.com/dbt-labs/dbt-core/blob/19c48e285ec381b7f7fa2dbaaa8d8361374136ba/core/dbt/config/runtime.py#L193-L203
    version_check = bool(flags.VERSION_CHECK)
    partial = Project.partial_load(project_root, verify_version=version_check)

    cli_vars: Dict[str, Any] = parse_cli_vars(getattr(args, "vars", "{}"))
    profile_renderer = ProfileRenderer(cli_vars)
    project_profile_name = partial.render_profile_name(profile_renderer)

    # from https://github.com/dbt-labs/dbt-core/blob/19c48e285ec381b7f7fa2dbaaa8d8361374136ba/core/dbt/config/profile.py#L423-L425
    profile_name = Profile.pick_profile_name(
        getattr(args, "profile", None), project_profile_name
    )
    raw_profiles = read_profile(flags.PROFILES_DIR)
    raw_profile = raw_profiles[profile_name]

    # from https://github.com/dbt-labs/dbt-core/blob/19c48e285ec381b7f7fa2dbaaa8d8361374136ba/core/dbt/config/profile.py#L287-L293
    target_override = getattr(args, "target", None)
    if target_override is not None:
        target_name = target_override
    elif "target" in raw_profile:
        target_name = profile_renderer.render_value(raw_profile["target"])
    else:
        target_name = "default"

    bv_dict = Profile._get_profile_data(raw_profile, profile_name, target_name)
    delegate = bv_dict.get("delegate")
    assert delegate, "buenavista targets must have a delegate field set"

    base_profile_dict = Profile.from_raw_profiles(
        raw_profiles,
        profile_name,
        renderer=profile_renderer,
        target_override=delegate,
    ).to_target_dict()

    bv_server = create(bv_dict, base_profile_dict)

    if bv_dict.get("api_port"):
        import threading
        import uvicorn
        from .api import app

        app.bv = bv_server
        bv_server_thread = threading.Thread(target=app.bv.serve_forever)
        bv_server_thread.daemon = True
        bv_server_thread.start()

        uvicorn.run(
            app,
            host=bv_dict.get("host"),
            port=bv_dict.get("api_port"),
            log_level=bv_dict.get("log_level", "info"),
        )
    else:
        print(f"Starting Buena Vista server at {bv_server.server_address}")
        bv_server.serve_forever()



#!/usr/bin/env python
import os
import re

from setuptools import find_namespace_packages
from setuptools import setup

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md")) as f:
    long_description = f.read()

package_name = "duckdbt"


package_version = "0.1.0"
description = """The Modern Data Stack in a Python Package"""

setup(
    name=package_name,
    version=package_version,
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Josh Wills",
    author_email="joshwills+duckdbt@gmail.com",
    url="https://github.com/jwills/duckdbt",
    packages=find_namespace_packages(include=["duckdbt", "duckdbt.*"]),
    include_package_data=True,
    install_requires=[
        "buenavista~=0.1.0",
        "dbt-buenavista~=1.3.0",
        "dbt-duckdb~=1.3.0",
        "duckdb~=0.6.0",
        "fastapi[all]",
        "pyarrow",
    ],
)
